""" Resolver interface for passing to processURI.

Resolvers are used to indirect the source and storage of templates and other
binary resources used by PXTL. The default StandardResolver accesses files and
other URIs directly, but other implementations of the interface could use
in-memory or database storage instead.
"""

__all__= [
  'Resolver',
  'ReadonlyResolver', 'StandardResolver', 'BytecodeResolver',
  'LookupResolver'
]

from pxtl.constants import *

import os, stat, string, time, urllib, urlparse


class Resolver:
  """ Interface for resolver-object arguments to processURI.

  If instanced directly, this class just returns file-not-found for everything
  and throws away any writes.
  """
  def getResource(self, uri, ctype):
    """ Redirect loading of resource indicated by given URI

    Implementation should return None if the resource could not be found,
    or a tuple of (resource, ctype, mtime), where resource is a byte stream,
    ctype is the MIME media type of the resource and mtime is the
    last-modified time as an integer timestamp. ctype/mtime can be None if
    unknown.

    The 'ctype' argument on entry gives the expected content type. This can be
    used to differentiate bytecode access from templates (and must be, since
    the URI passed in is the same for both cases).
    """
    return None

  def setResource(self, uri, ctype, content):
    """ Write binary content to resource

    Currently 'ctype' will always be bytecode. The method should return True
    if the resource was successfully written or False otherwise, though this
    is not currently actually used.
    """
    return False

  def makeCompiledURI(self, uri):
    """ Return a URI with a filename extension for compiled bytecode

    Utility function for subclasses that want to store bytecode in eg. .pxc
    files next to .px files.
    """
    uriparts= list(urlparse.urlparse(uri))
    leaf= string.split(uriparts[2], '/')[-1]
    if '.' in leaf and string.lower(string.split(leaf, '.')[-1])=='px':
      uriparts[2]= uriparts[2]+'c'
    else:
      uriparts[2]= uriparts[2]+'.pxc'
    return urlparse.urlunparse(uriparts)


class ReadonlyResolver(Resolver):
  """ Standard reading-resolver with no write capability

  Uses urllib to retrieve URIs (generally from the filesystem)
  """
  def getResource(self, uri, ctype):
    if ctype in TYPES_BYTECODE:
      uri= self.makeCompiledURI(uri)

    # Provide separate processing for file: URIs. This is (a) to simplify the
    # common case, (b) because using urllib to fetch a file doesn't return a
    # Last-Modified time header on older versions of Python, and (c) because
    # urllib/mimetypes unhelpfully guesses 'text/plain' for unknown file
    # extensions, where we'd rather default to .px type.
    #
    uriparts= urlparse.urlparse(uri)
    if string.lower(uriparts[0])=='file':
      path= urllib.url2pathname(uriparts[2])
      ext= os.path.splitext(os.path.basename(path))[1]
      try:
        mtime= os.stat(path)[stat.ST_MTIME]
        ctype= mimetypes.types_map.get(ext, None)
        fp= open(path, 'rb')
        return (fp, ctype, mtime)
      except (OSError, IOError): return None

    # For other URI types let urllib try to give us workable access.
    #
    try:
      fp= urllib.urlopen(uri)
      mtime= parseFullTime(fp.info().get('Last-Modified', ''))
      ctype= fp.info().get('Content-Type', None)
      return (fp, ctype, mtime)
    except (OSError, IOError): return None


class StandardResolver(ReadonlyResolver):
  """ Default URI/file-based implementation of PXTL Resolver interface

  Uses the same read code as ReadonlyResolver, adds writing capability
  """
  def setResource(self, uri, ctype, content):
    if ctype in TYPES_BYTECODE:
      uri= self.makeCompiledURI(uri)

    # Attempt to store resources only for file: URIs... no point in trying to
    # HTTP PUT or anything slow like that
    #
    uriparts= urlparse.urlparse(uri)
    if uriparts[0].lower()=='file':
      path= urllib.url2pathname(uriparts[2])
      try:
        if not os.path.exists(os.path.dirname(path)):
          os.makedirs(os.path.dirname(path))
        fp= open(path, 'wb')
        fp.write(content)
        fp.close()
      except (OSError, IOError):
        pass
      else:
        return True
    return False


class BytecodeResolver(StandardResolver):
  """ Resolver that can fetch bytecode from a different location to templates

  A lookup from source template stem keys to bytecode stem values is kept.
  When a URI is resolved, the most closely-matching template stem is chosen
  and that part of the path replaced with the corresponding bytecode stem.
  This allows bytecode files to be stored in different folders to the template
  files, depending on their location.
  """
  def __init__(self,
    bytecodebase, isfilebased= False, casesensitive= True, _exactpath= False
  ):
    """ Initialise with a lookup dictionary of template to bytecode stems

    The stems must be of absolute URIs by default. If you set 'isfilebased',
    they are instead possibly-relative filesystem paths, which will be
    converted to absolute URIs at initialisation time.

    'casesensitive' (on by default) requires that stems match exactly. If
    unset, the case doesn't matter on matching (but will be preserved on
    translation).

    If 'bytecodebase' is None, no bytecode files will ever be used.
    """
    # _exactpath can be set True as a hack to disable the automatic adding of
    # 'c' to bytecode filenames. This is used by the legacy compileFile
    # interface.

    # Convert file-based lookup to plain URI lookup
    #
    if isfilebased and bytecodebase is not None:
      uricodebase= {}
      for tstem, bstem in bytecodebase.items():
        tstem= 'file://'+urllib.pathname2url(os.path.abspath(tstem))
        bstem= 'file://'+urllib.pathname2url(os.path.abspath(bstem))
        uricodebase[tstem]= bstem
        bytecodebase= uricodebase

    # Pre-lowercase keys if case-insensitive
    #
    if not casesensitive and bytecodebase is not None:
      cicodebase= {}
      for tstem, bstem in bytecodebase.items():
        cicodebase[tstem.lower()]= bstem
      bytecodebase= cicodebase

    # Store for translateURI
    #
    self.bytecodebase= bytecodebase
    self.casesensitive= casesensitive
    self.exactpath= _exactpath

  # Funnel get/set through translateURI
  #
  def getResource(self, uri, ctype):
    if ctype in TYPES_BYTECODE:
      uri= self.translateURI(uri)
    if uri is None:
      return None
    return StandardResolver.getResource(self, uri, ctype)

  def setResource(self, uri, ctype, content):
    if ctype==TYPES_BYTECODE[0]:
      uri= self.translateURI(uri)
    if uri is None:
      return False
    return StandardResolver.setResource(self, uri, ctype, content)

  def makeCompiledURI(self, uri):
    if self.exactpath:
      return uri
    return StandardResolver.makeCompiledURI(self, uri)

  def translateURI(self, uri):
    """ Return URI translated through the lookup supplied to initialiser

    Returns the URI with its most-deeply-matching template stem replaced by
    the corresponding bytecode stem, or the same URI if no match was found.
    Returns None if bytecode should not be used at all.
    """
    if self.bytecodebase is None:
      return None

    urici= uri
    if not self.casesensitive:
      urici= uri.lower()

    # Search keys for match in descending order of length, to catch most-
    # matching first. Require a URL-part terminator or end-of-string following
    # the match in order to prevent eg. folder 'foo' also matching 'foo.px'.
    #
    keys= self.bytecodebase.keys()
    keys.sort(lambda a, b: -cmp(len(a), len(b)))
    for key in keys:
      keylen= len(key)
      if urici[:keylen]==key and urici[keylen:keylen+1] in ('', '/', '?', ';'):
        return self.bytecodebase[key]+uri[keylen:]
    return uri


class LookupResolver(Resolver):
  """ In-memory storage using a dictionary of {uri: (content, ctype, mtime)}

  If 'mutable' (default yes), allow resources to be stored back in the lookup
  else read-only.
  """
  def __init__(self, lookup, mutable= True):
    self.lookup= lookup
    self.mutable= mutable

  def getResource(self, uri, ctype):
    if ctype in TYPES_BYTECODE:
      uri= self.makeCompiledURI(uri)
    if not self.lookup.has_key(uri):
      return None
    content, ctype, mtime= self.lookup[uri]
    return (StringIO(content), ctype, mtime)

  def setResource(self, uri, ctype, content):
    if ctype in TYPES_BYTECODE:
      uri= self.makeCompiledURI(uri)
    if self.mutable:
      self.lookup[uri]= (content, ctype, time.time())
      return True
    return False


# Consider adding database-backed SQLResolver?


# Utility
#
fulltime_MONTHS= {
  'jan': 1, 'january': 1, 'feb': 2, 'february': 2, 'mar': 3, 'march': 3,
  'apr': 4, 'april': 4, 'may': 5, 'jun': 6, 'june': 6, 'jul': 7, 'july': 7,
  'aug': 8, 'august': 8, 'sep': 9, 'september': 9, 'oct': 10, 'october': 10,
  'nov': 11, 'november': 11, 'dec': 12, 'december': 12
}
DAY= 60*60*24
def parseFullTime(s):
  """ Convert Last-Modified time string format to timestamp

  Utility function for the StandardResolver. Returns None if the time is
  unreadable. See RFC2616 3.1.3 for details of this stupid format. A fuller
  (stricter) parser is included in the email module in later Python versions.
  """
  s= string.replace(string.replace(string.lower(s), '-', ' '), ':', ' ')
  words= string.split(s, ' ')
  if len(words)<7:
    return None

  # asctime format (ugh)
  #
  if fulltime_MONTHS.has_key(words[1]):
    month= fulltime_MONTHS[words[1]]
    try:
      day, hour, minute, second, year= map(int, words[2:7])
    except ValueError: return None

  # RFC822 or 850 format (either should work)
  #
  else:
    try:
      month= fulltime_MONTHS[words[2]]
    except KeyError: return None
    try:
      day= int(words[1])
      year, hour, minute, second= map(int, words[3:7])
    except ValueError: return None

  # Check ranges (fairly liberally)
  #
  if 0<=year<100:
    year= year+2000
  if not (
    year>=0 and 1<=day<=31 and 0<=hour<=24 and 0<=minute<=59 and 0<=second<=62
  ): return None

  # Make GMT timestamp (roundabout way due to lack of time.mkgmttime)
  #
  t= (year, month, day, hour, minute, second, 0, 0, 0)
  local= time.mktime(t)
  epoch= time.mktime((1970, 1, 3, 0, 0, 0, 3, 1, t[8]))-DAY*2
  return local-epoch

